/*
			 Salvatore Ruggieri (c), 2008
*/

/** @file clpt.cpp */

#include "poly.h"

/** Well-mode checking procedure */
long leftcheck(Program &rules, bool forceModes, string forceMode, bool lpflag, bool ieflag, 
					bool poflag, bool paflag, bool lp2flag)
{
	// output modes
	size_t nm = rules.modes.size();
	for(size_t i=0; i<nm; ++i)
		cout << "mode: " << *(rules.modes[i]) << endl;
	cout << "--" << endl;
	// no rules
	size_t n = rules.rules.size();
	// rules heads
	long atoms = n;
	// force moding	
	Type modeforce( forceModes ? forceMode : "!" );
	bool someflag = lpflag || ieflag || poflag || paflag || lp2flag;
	if( !someflag && forceModes ) {
		someflag = true;
		Type::TypeSystem ts = modeforce.typeSystem();
		lpflag = ts == Type::UP2BOX;
		ieflag = ts == Type::UP2DEF;
		poflag = ts == Type::UP2BT;
		paflag = ts == Type::UP2PT;
	}
	for(size_t i=0; i<n; ++i) {
		// is well-moded?
		bool wm = true;
		// output rule
		const Rule *r = rules.rules[i];
		const string &c = r->getConstraint();
		cout << "rule: " // << i << " of " << n << " " 
						<< *r << endl;
		if( !someflag ) {
			Type::TypeSystem ts = r->typeSystem(rules.modes);
			lpflag = ts == Type::UP2BOX;
			ieflag = ts == Type::UP2DEF;
			poflag = ts == Type::UP2BT;
			paflag = ts == Type::UP2PT;
		}
	    ISolve *poly = NULL;
		if( lpflag )
			poly = (ISolve*)(new LPCheck());
		else if( lp2flag )
			poly = (ISolve*)(new LP2Check());
		else if( ieflag )
			poly = (ISolve*)(new IECheck());
		else if( poflag )
			poly = (ISolve*)(new POLYCheck()); 
		else if( paflag ) {
			// computes longest type declaration (pre(HEAD), post(BODY)) to estimate parameter pool optimal size
			// such a computation is not strictly necessary, but
			// non-optimal pool size leads to: inefficiency (if overestimated), out of parameter pool (if underestimated)
			const Atom *head = r->getHead();
			Mode *headmode = head->findMode(rules.modes);
			if( (headmode == NULL) && !forceModes )
				throw new runtime_error("Error: " + head->getPredicate() + " has no mode.");
			string headin = forceModes ? head->getModes(modeforce) : head->getInputModes(headmode);
			TypeDecl td(headin);
			if( !forceModes ) {
				// foreach body atom
				size_t len = r->size();
				for(size_t j=0; j<len; ++j) {
					const Atom *body = r->getBodyAtom(j);
					Mode *bodymode = body->findMode(rules.modes);
					if( bodymode == NULL )
						throw new runtime_error("Error: " + body->getPredicate() + " has no mode.");
					td.add(body->getOutputModes(bodymode));
				}
			}
			ParCheck *p = new ParCheck();
			p->set_par_pool(td,c);
			poly = (ISolve*)p;
		}
		// set constraint
        poly->set(c);
		if( !poly->isCFeasible() ) {
			cout << "--> Well-moded. But inconsistent constraint." << endl;
			delete poly;
			continue;
		}
		// head
		const Atom *head = r->getHead();
		Mode *headmode = head->findMode(rules.modes);
		if( (headmode == NULL) && !forceModes )
			throw new runtime_error("Error: " + head->getPredicate() + " has no mode.");
		string headin = forceModes ? head->getModes(modeforce) : head->getInputModes(headmode);
		TypeDecl td(headin);
		// foreach body atom
		size_t len = r->size();
		atoms += len;
		for(size_t j=0; j<len; ++j) {
			const Atom *body = r->getBodyAtom(j);
			Mode *bodymode = body->findMode(rules.modes);
			if( (bodymode == NULL) && !forceModes )
				throw new runtime_error("Error: " + body->getPredicate() + " has no mode.");
			string bodyin = forceModes ? body->getModes(modeforce) : body->getInputModes(bodymode);
			TypeDecl td_toshow(bodyin);
			string diff;
			if( !poly->check(td, td_toshow, diff) ) {
				cout << "  at step = " << j << endl;
				cout << "  type decl = " << td << endl;
				cout << "  error: " << diff << endl;
				wm = false;
			}
			if( forceModes )
				continue; // nothing to add
			string bodyout = body->getOutputModes(bodymode);
			td.add(bodyout);
		}
		// mode proof obligation for head output modes
		if( !forceModes ) {
			string headout = head->getOutputModes(headmode);
			string diff;
			TypeDecl td_toshow(headout);
			if( !poly->check(td, td_toshow, diff) ) {
				cout << "  at step = " << len << endl;
				cout << "  type decl = " << td << endl;
				cout << "  error: " << diff << endl;
				wm = false;
			}
		}
		if( wm )
			cout << "--> OK well-moded." << endl;
		else
			cout << "--> Not well-moded." << endl;
		delete poly;
	}
	return atoms;
}

/** Test a type assertion checking procedure. */
void test(ISolve &check, const string &constraint, const TypeDecl &td, const TypeDecl &td2)
{
	cronos t;
	t.start();
	check.set(constraint);
	t.stop();
	cout << "  feasible = " << (check.isCFeasible() ? "yes" : "no") <<
		" (" << t.get_elapsed() << " secs)." << endl;
	t.start();
	cout << "  valid = ";
	string diff;
	if( check.check(td, td2, diff) )
		cout << "yes";
	else {
		if( !check.is_check_complete(td, td2) )
			cout << "maybe";
		else
			cout << "no";
		cout << " [" << diff << "]";
	}
	t.stop();
	cout << " (" << t.get_elapsed() << " secs)." << endl;
}

/** Test a type assertion inference procedure. */
void test(ISolve &infer, const string &constraint, const TypeDecl &td)
{
	cronos t;
	t.start();
	infer.set(constraint);
	t.stop();
	cout << "  feasible = " << (infer.isCFeasible() ? "yes" : "no") <<
		" (" << t.get_elapsed() << " secs)." << endl;
	t.start();
	TypeDecl *inferred = infer.infer(td);
	cout << "  output = " << inferred->to_string();
	delete inferred;
	t.stop();
	cout << " (" << t.get_elapsed() << " secs)." << endl;
}

const string &version = "CLPT - Typing linear constraints - Version 1.3 (November 2008)";
const string &copyright = "Copyright (C) 2008 Salvatore Ruggieri (ruggieri@di.unipi.it)";

/** Program usage. */
void usage()
{
	cout << version << endl;
	cout << copyright << endl;
	cout << "Usage: clpt [-h | -l | -v | -p[ -m<m> ] | -lp | -lp2 | -ie | -po | pa] < <assertion_or_program_file>" << endl;
	cout << "  -h help in usage" << endl;
	cout << "  -l licence" << endl;
	cout << "  -v verbose mode" << endl;
	cout << "  -p CLP(R) well-mode checking using ParCheck" << endl;
	cout << "  -m<m> all modes forced to: <m> x <m>" << endl;
	cout << "  -lp use Check(LPCheck) (restricts to BT_Box)" << endl;
	cout << "  -lp2 use Check(LP2Check) (restricts to BT_Box)" << endl;
	cout << "  -ie use Check(IECheck) (restricts to BT_!)" << endl;
	cout << "  -po use Check(PolyCheck) (restricts to BT)" << endl;
	cout << "  -pa use ParCheck (default)" << endl;
	cout << "If -p is specified, then at most one of  -lp -lp2 -ie -po is admitted." << endl;
	cout << "Sample usages:" << endl;
	cout << "clpt < assertions/infer/paper_ex1.query" << endl;
	cout << "clpt -po -ie < assertions/check/iecheck_may_polycheck_no.query" << endl;
	cout << "clpt -p < programs/folk/ack1.clpq" << endl;
	cout << "clpt -p -m! < programs/folk/ack1.clpq" << endl << endl;
	cout << "Type syntax:" << endl;
	cout << "|e  upper and lower bound wrt parametric expression e" << endl;
	cout << "|r  upper and lower bound of width at most r" << endl;
	cout << "!  definiteness" << endl;
	cout << "|  upper and lower bound" << endl;
	cout << "^  upper bound" << endl;
	cout << "_  lower bound" << endl;
	cout << "*  no bound" << endl;
	exit(0);
}

void licence()
{
	cout << version << endl;
	cout << copyright << endl << endl;
	
    cout << "This program is free software: you can redistribute it and/or modify" << endl;
    cout << "it under the terms of the GNU General Public License as published by" << endl;
    cout << "the Free Software Foundation, either version 3 of the License, or" << endl;
    cout << "(at your option) any later version." << endl << endl;

    cout << "This program is distributed in the hope that it will be useful," << endl;
    cout << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl;
    cout << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << endl;
    cout << "GNU General Public License for more details." << endl;

    cout << "You should have received a copy of the GNU General Public License" << endl;
    cout << "along with this program.  If not, see <http://www.gnu.org/licenses/>." << endl;
	exit(0);
}

/** Main program. */
int main(int argc, char *argv[])
{
	// program parameters
	bool program = false;
	bool forceModes = false;
	string forceMode;
	bool lpflag = false;
	bool lp2flag = false;
	bool ieflag = false;
	bool poflag = false;
	bool paflag = false;

	if( argc > 1  ) {
		for(int i=1; i<argc; ++i ) {
			if( strcmp(argv[i],"-v")==0 )
				verbose::level = 1;
			else if( strcmp(argv[i],"-p")==0 )
				program = true;
			else if( strcmp(argv[i],"-lp")==0 )
				lpflag = true;
			else if( strcmp(argv[i],"-lp2")==0 )
				lp2flag = true;
			else if( strcmp(argv[i],"-ie")==0 )
				ieflag = true;
			else if( strcmp(argv[i],"-po")==0 )
				poflag= true;
			else if( strcmp(argv[i],"-pa")==0 )
				paflag= true;
			else if( strcmp(argv[i],"-l")==0 )
				licence();
			else if( strncmp(argv[i],"-m",2)==0 ) {
				forceModes = true;
				forceMode = string(&argv[i][2]);
			}
			else usage();
		}
		int somemode = lpflag + ieflag + poflag + lp2flag + paflag;
		if( program && somemode > 1 )
			usage();
		if( !program && forceModes )
			usage();
	}
	try {

		if( program )
		{
			// parsing
			cronos t;
			t.start();
			Program pgr( Program::parse(cin) );
			t.stop();
			cout << "Parsing: " << t.get_elapsed() << " secs." << endl;
			// checking well-moding
			t.start();
			long atoms = leftcheck(pgr,forceModes,forceMode,lpflag,ieflag,poflag,paflag,lp2flag);
			t.stop();
			cout << "Well-mode checking: " << t.get_elapsed() << " secs." << endl;
			cout << "No clauses: " << pgr.rules.size() << endl;
			cout << "Avg time per clause: " << t.get_elapsed()/pgr.rules.size() << " secs." << endl;
			cout << "No atoms: " << atoms << endl;
			cout << "Avg time per atom: " << t.get_elapsed()/atoms << " secs." << endl;
			return 0;
		}

		// read constraint and type declaration from stdin
		string constraint, typedecl, expected;
		getline(cin, constraint);
		size_t l = constraint.size();
		// remove trail \r (Windows editors)
		if( l > 0 && constraint[l-1] == '\r' )
			constraint.resize(l-1);
		getline(cin, typedecl);
		l = typedecl.size();
		if( l > 0 && typedecl[l-1] == '\r' )
			typedecl.resize(l-1);
		if( !cin.eof() ) {
			getline(cin, expected);
			l = expected.size();
			if( l > 0 && expected[l-1] == '\r' )
				expected.resize(l-1);
		}

		TypeDecl td(typedecl);
		cout << "input: " << constraint << endl;
		cout << "decl: " << td << endl;
		
		if( lpflag + ieflag + poflag + paflag + lp2flag == 0 )
			lpflag = ieflag = poflag = paflag = lp2flag = true;
		
		if( expected == "" ) {
			// inference
			if( lpflag ) {
				LPCheck check;
				cout << "LPInfer" << endl;
				test(check, constraint, td);
			}
			if( lp2flag ) {
				LP2Check check;
				cout << "LP2Infer" << endl;
				test(check, constraint, td);
			}
			if( ieflag ) {
				IECheck check;
				cout << "IEInfer" << endl;
				test(check, constraint, td);
			}
			if( poflag ) {
				POLYCheck check;
				cout << "POLYInfer" << endl;
				test(check, constraint, td);
			}
		} else {
			// checking
			TypeDecl td2(expected);
			cout << "expd: " << td2 << endl;
			if( lpflag ) {
				LPCheck check;
				cout << "Check(LPInfer)" << endl;
				test(check, constraint, td, td2);
			}
			if( lp2flag ) {
				LP2Check check;
				cout << "Check(LP2Infer)" << endl;
				test(check, constraint, td, td2);
			}
			if( ieflag ) {
				IECheck check;
				cout << "Check(IECheck)" << endl;
				test(check, constraint, td, td2);
			}
			if( poflag ) {
				POLYCheck check;
				cout << "Check(POLYInfer)" << endl;
				test(check, constraint, td, td2);
			}
			if( paflag ) {
				ParCheck check;
				cout << "ParCheck" << endl;
				test(check, constraint, td, td2);
			}
		}

	} catch(runtime_error e) {
		cout << e.what() << endl;
		cout.flush();
	}
	return 0;
}
