/*
			 Salvatore Ruggieri (c), 2008
*/

/** @file poly.h  */

#ifndef _POLY_H__INCLUDED_
#define _POLY_H__INCLUDED_

#include "basic.h"

/** The linear constraint typing. */
namespace clpt
{
/** Abstract class for type inference and checking. */
class ISolve
{
public:
	/** Destructor. */
	virtual ~ISolve() {}

	/** Set the reference constraint. */
	virtual void set(const string &constraint) = 0;

	/** Is the reference constraint feasible?. */
	virtual bool isCFeasible() const = 0;

	/** Get the list of variables in the reference constraint. */
	virtual string getCVarList() const = 0;

	/** Inference for a given type declaration with respect to a given list of variables. */
	virtual TypeDecl *infer(const TypeDecl &typedecl, const string &varlist) = 0;

	/** Inference for a given type declaration with respect to all variables. */
	virtual TypeDecl *infer(const TypeDecl &typedecl)
	{
		string vars = typedecl.getVarList();
		string cvars = getCVarList();
		if( cvars == "" )
			return infer(typedecl, vars);
		if( vars == "" )
			return infer(typedecl, cvars);
		vars.append(",");
		vars.append(cvars);
		return infer(typedecl, vars);
	}

	/** Checking a type assertion. Use the infer() function. */
	virtual bool check(const TypeDecl &td1, const TypeDecl &td2, string &diff)
	{
		if( verbose::level > 0 ) {
			cout << "Calling checking procedure" << endl;
			cout << "td1 = " << td1.to_string() << endl;
			cout << "td2 = " << td2.to_string() << endl;
			cout << "vars = " << td2.getVarList() << endl;
		}
		const TypeDecl *inferred = infer(td1, td2.getVarList());
		bool result = inferred->implies(&td2);
		if( !result )
			diff = "expected " + td2.to_string() + " - actual: " + inferred->to_string();
		return result;
	}

	/** Return true if checking is complete for the provided type declarations. */
	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const = 0;
};


/** Implements the %LPInfer procedure using Minkowski's decomposition. */
class LPCheck : public ISolve
{
public:
	/** Constructor. */
	LPCheck()
	: ls(NULL), polHomo(NULL), polNonHomo(NULL), feasible(true)
	{
	}

	/** Destructor. */
	virtual ~LPCheck()
	{
		clear();
	}

	virtual void set(const string &constraint)
	{
		clear();
		// convert constraint to a linear system
		ls = new LinearSystem(constraint);
		if( verbose::level > 0 ) {
			cout << "LPCheck called" << endl;
			cout << "Constraint: " << constraint << endl;
			cout << "Linear system:" << endl << *ls << endl;
		}
		// get variable list
		cvars = ls->getVarList();
		nvars_c = ls->getNcols();

		_duplicatevars(ls);
		// convert linear system to Polylib Matrix
		Matrix *lsm = ls->toPolylib();
		if( verbose::level > 0 ) {
			cout << "Matrix:" << endl;
			Matrix_Print(stdout,P_VALUE_FMT,lsm);
		}
		// and next to a Polyhedron
		cronos c;
//		c.start();
		polNonHomo = Constraints2Polyhedron(lsm,0);
		Matrix_Free(lsm);
//		c.stop();
//		cout << "Satisfiability check. Elapsed time = " << c.get_elapsed() << endl;
		if( verbose::level > 0 ) {
			cout << "Polyhedron:" << endl;
			Polyhedron_Print(stdout,P_VALUE_FMT,polNonHomo);
		}
		// check if Polyhedron is empty
		feasible = polNonHomo->NbRays != 0;
		if( feasible && ls->isStrict() ) {
			// non-strict feasible, check if strict is feasible
			size_t n_rows = ls->getNrows();
			Matrix *lsm = ls->toPolylib(); // needed to regenerate, seems Constraints2Polyhedron overwrites it
			for (size_t i = 0 ; i < n_rows; i++)
				if( (ls->getRowType(i) == LinearSystem::MAT_LT) &&
					poly_util::expr_equal(polNonHomo, lsm->p[i]) ) {
						feasible = false;
						break;
					}
			Matrix_Free(lsm);
		}
		//
		if( !feasible ) {
			// empty polyhedron
			if( verbose::level > 0 )
				cout << "Empty polyhedron." << endl;
			delete ls;
			ls = NULL;
			Domain_Free(polNonHomo);
			polNonHomo = NULL;
			return;
		}
		if( ls->isHomo() )
			polHomo = Polyhedron_Copy(polNonHomo);
		else {
			// homogeneous matrix
			Matrix *lsmh = ls->toPolylibHomo();
			if( verbose::level > 0 ) {
				cout << "Matrix Homo:" << endl;
				Matrix_Print(stdout,P_VALUE_FMT,lsmh);
			}		
			// build Polyhedron for the homogeneous system
			polHomo = Constraints2Polyhedron(lsmh,0);
			Matrix_Free(lsmh);
		}
		if( verbose::level > 0 ) {
			cout << "Polyhedron Homo:" << endl;
			Polyhedron_Print(stdout,P_VALUE_FMT,polHomo);
		}
	}

	virtual bool isCFeasible() const
	{
		return feasible;
	}

	virtual string getCVarList() const
	{
		return cvars;
	}

	/** Get the list of variables in the reference constraint. */
	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const
	{
		if( !feasible )
			return true;
		if( td1.typeSystem() > Type::UP2BT ) // parametric types
			return false;
		Type::TypeSystem ts2 = td2.typeSystem();
		if( ts2 <= Type::UP2BOX )
			return true;
		if( ts2 == Type::UP2BT )
			return ls->getNrows()==0;
		return false;
	}

	virtual TypeDecl *infer(const TypeDecl &typedecl, const string &varlist)
	{
		if( verbose::level > 0 ) {
			cout << "LPCheck::infer called" << endl;
			cout << "Type declaration: " << typedecl << endl;
			cout << "Variables to infer:" << varlist << endl;
		}
		// assign at least the types in typedecl, remove duplicates in varlist
		TypeDecl *v = typedecl.restrict(varlist);
		// unsatisfiable constraint first
		if( !feasible ) {
			v->setAllType(DEF);
			if( verbose::level > 0 ) {
				cout << "Unsatisfiable constraint." << endl;
				cout << "Inferred type declaration: " << v->to_string() << endl;
			}
			return v;
		}
		// check if needed to perform LPInfer-ence
		size_t sz = v->size();
		size_t n_cols = ls->getNcols();
		// count variables that have to be typed
		size_t to_type = 0;
		for(size_t i=0; i<sz; ++i)
			if( !(v->getType(i) >= BOX) && // x : \tau in typedecl, with BOX >_t \tau
				(ls->getColPos(v->getVarName(i)) < n_cols) ) // x in vars(c)
				++to_type;
		if( to_type == 0 )
			return v;

		// Polylib Matrix for homogeneous phi(typedecl)
		Matrix *tdm = _get_phi(typedecl, ls);
		if( verbose::level > 0 ) {
			cout << "Homogeneous phi(typedecl):" << endl;
			Matrix_Print(stdout,P_VALUE_FMT,tdm);
		}
		// and its polyhedron
		Polyhedron *tdpol = Constraints2Polyhedron(tdm,0);
		// compute Sol(c) \intersect Sol(phi(typedecl))
		Polyhedron *pol = DomainIntersection(polHomo, tdpol, 0);
		Matrix_Free(tdm);
		Domain_Free(tdpol);

		if( verbose::level > 0 ) {
			cout << "Homogeneous polyhedron:" << endl;
			Polyhedron_Print(stdout,P_VALUE_FMT,pol);
		}

		// for each variable to be typed
		for(size_t i=0; i<sz; ++i)
			if( !(v->getType(i) >= BOX) ) {
				size_t idx = ls->getColPos(v->getVarName(i));
				if(idx < n_cols) // x in vars(c)
					v->setType(i, _infer(pol, idx, v->getType(i).tp)); // compute type
			}
		Domain_Free(pol);

		if( verbose::level > 0 )
			cout << "Inferred Type declaration: " << v->to_string() << endl;
		return v;
	}

protected:
	/** Constraint in linear system form. */
	LinearSystem *ls;
	/** Polyhedron for the homogeneous system. */
	Polyhedron *polHomo;
	/** Polyhedron for the non-homogeneous system. */
	Polyhedron *polNonHomo;
	/** Is constraint satisfiable? */
	bool feasible;
	/** Constraint variables. */
	string cvars;
	/** No of constraint variables. */
	size_t nvars_c;

	/** Clear data. */
	virtual void clear()
	{
		if( polNonHomo == polHomo )
			polNonHomo = NULL;
		if( polHomo != NULL ) {
			Domain_Free(polHomo);
			polHomo = NULL;
		}
		if( polNonHomo != NULL ) {
			Domain_Free(polNonHomo);
			polNonHomo = NULL;
		}
		if( ls != NULL ) {
			delete ls;
			ls = NULL;
		}
		feasible = true;
		cvars = "";
	}

	/** Stub method. Needed in POLYCheck. */
	virtual void _duplicatevars(LinearSystem *m)
	{
	}

	/** Returns the matrix for homogeneous phi(td). */
	static Matrix * _get_phi(const TypeDecl &td, LinearSystem *m)
	{
		size_t ncol = m->getNcols();
		size_t sz = td.size();
		// count rows
		size_t nrows = 0;
		for(size_t i=0; i<sz; ++i)
			if( (td.getType(i) != STAR) &&
				(m->getColPos( td.getVarName(i) ) < ncol) )
				++nrows;
		Matrix *M = Matrix_Alloc(nrows, ncol+2);
		// set rows
		size_t rowp = 0;
		for(size_t i=0; i<sz; ++i)
		{
			if( td.getType(i) == STAR ) continue;
			size_t pos = m->getColPos( td.getVarName(i) ); // x not in vars(c)
			if( pos >= ncol ) continue;
			::Value *row = M->p[rowp++];
			switch( td.getType(i).tp )
			{
				case DEF:
				case BOX:
					row[0] = 0;
					row[pos+1] = 1;
					row[ncol+1] = 0;
					break;
				case UP:
					row[0] = 1;
					row[pos+1] = -1;
					row[ncol+1] = 0;
					break;
				case DOWN:
					row[0] = 1;
					row[pos+1] = 1;
					row[ncol+1] = 0;
					break;
				default:
					break;
			}
		}
		return M;
	}

	/** Infer the type of a variable given a polyhedron. */
	Type _infer(Polyhedron *Pol, size_t idx, Type mintp)
	{
		size_t Dimension = Pol->Dimension;
		size_t NbRays = Pol->NbRays;
		Type val = BOX;
		for (size_t i=0;i<NbRays;++i) {
			::Value *p=Pol->Ray[i];
			if( value_notzero_p(*p) ) {
				::Value pd1 = p[Dimension+1];
				if(value_notzero_p(pd1) )  // vertex
					continue;
				// ray
				::Value v = p[idx+1];
				if( v > 0 ) {
					if( val == BOX ) {
						val = DOWN;
						if( val == mintp )
							return val;
					}
					else if( val == UP )
							return STAR;
				}
				else if( v < 0 ) {
					if( val == BOX ) {
						val = UP;
						if( val == mintp )
							return val;
					}
					else if( val == DOWN )
							return STAR;
				}
			} else { // line
				if( p[idx+1] != 0 )
					return STAR;
			}
		}
		return val;
	}
};


/** Implements the %IEInfer procedure. */
class IECheck : public LPCheck
{
public:
	/** Constructor. */
	IECheck()
	: LPCheck()
	{
	}

	/** Destructor. */
	virtual ~IECheck()
	{
	}

	virtual TypeDecl *infer(const TypeDecl &typedecl, const string &varlist)
	{
		if( verbose::level > 0 ) {
			cout << "IECheck::infer called" << endl;
			cout << "Type declaration: " << typedecl << endl;
			cout << "Variables to infer:" << varlist << endl;
		}
		TypeDecl *v = LPCheck::infer(typedecl, varlist);
		// unsatisfiable or true constraint, LPCheck is complete
		if( !feasible )
			return v;
		// check if candidate exists for IEInfer-ence
		size_t sz = v->size();
		bool gonext = false;
		size_t n_cols = ls->getNcols();
		for(size_t i=0; i<sz; ++i)
			if( (v->getType(i) == BOX) && (ls->getColPos(v->getVarName(i)) < n_cols) )
				{
					gonext = true;
					break;
				}
		if( !gonext )
			return v;

		// move def-typed variables in higher dimensions, both in ls, polHomo and polNonHomo
		// this is equivalent to introduce explicit parameters
		// and equations x = a for x:! in typedecl.
		// return count of number of def-typed variables in typedecl
		size_t nparams = _arrange_def(typedecl);
		_infer(typedecl, nparams, v);
		_unarrange_def();

		if( verbose::level > 0 )
			cout << "Inferred Type declaration: " << v->to_string() << endl;
		return v;
	}

	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const
	{
		if( !feasible )
			return true;
		if( td1.typeSystem() > Type::UP2BT ) // parametric types
			return false;
		Type::TypeSystem ts2 = td2.typeSystem();
		if( ts2 <= Type::UP2DEF )
			return true;
		if( ts2 == Type::UP2BT )
			return ls->getNrows()==0;
		return false;
	}

protected:

    vector<bool> arranged;

    void _unarrange_def()
	{
		size_t n_cols=ls->getNcols();
		size_t last = n_cols;
		for(size_t i=0; i<n_cols; ++i)
			if( arranged[i] ) {
				ls->swap_col(i, --last);
				poly_util::swap_pol_dim(polNonHomo, i, last);
				poly_util::swap_pol_dim(polHomo, i, last);
			}
	}

	/** Put def-typed variables at the higher dimensions. */
	size_t _arrange_def(const TypeDecl &typedecl)
	{
		size_t n_cols = ls->getNcols();
		size_t last = n_cols;
		arranged.resize(n_cols);
		fill(arranged.begin(), arranged.end(), false);

		// count no of defs in input type declaration
		size_t sztd = typedecl.size();
		size_t def_typed = 0;
		for(size_t i=0; i<sztd; ++i) {
			const Type &type = typedecl.getType(i);
			if( type == DEF) {
				const string &varname = typedecl.getVarName(i);
				size_t colpos = ls->getColPos(varname);
				if( colpos >= n_cols )
					continue;
				def_typed++;
				ls->swap_col(colpos, --last);
				poly_util::swap_pol_dim(polNonHomo, colpos, last);
				poly_util::swap_pol_dim(polHomo, colpos, last);
				arranged[colpos] = true;
			}
		}
		return def_typed;
	}


	/** Inference. */
	void _infer(const TypeDecl &typedecl, size_t nparams, TypeDecl *v)
	{
		size_t n_cols = ls->getNcols();
		size_t sz = v->size();

		Matrix mat;
		mat.p = polNonHomo->Constraint;
		mat.NbRows = polNonHomo->NbConstraints;
		unsigned Neq = polNonHomo->NbEq;
		Gauss(&mat, Neq, polNonHomo->Dimension);

		if( verbose::level > 0 ) {
			  cout << "No of parameters = " << nparams << endl;
			  char **param_name = new char*[nparams];
			  cout << "Parameters:" << endl;
			  for(size_t i =0; i < nparams; ++i ) {
				  param_name[i] = (char *)ls->getColName(n_cols - nparams + i).c_str();
				  cout << "  " << param_name[i] << endl;
			  }
			  cout << "NonHomo polyhedron:" << endl;
			  Polyhedron_Print(stdout,P_VALUE_FMT,polNonHomo);
			  delete[] param_name;
		}

		unsigned nVar = polNonHomo->Dimension - nparams;

		for(size_t i=0; i<sz; ++i) {
			if( v->getType(i) != BOX )
				continue;
			size_t idx = ls->getColPos(v->getVarName(i));
			if( idx >= n_cols )
				continue;
//				cout << "Checking: " << v->getVarName(i) << " at position: " << idx << endl;
			if(	check_implicit(polNonHomo->Constraint, Neq, nVar, idx) )
				v->setType(i, DEF);
		}
	}

	/** Auxiliary method. */
	static bool check_implicit(Value **Constraint, size_t Neq, size_t nVar, size_t idx)
	{
		for(size_t i=0; i<Neq; ++i) {
			Value *row = Constraint[i];
			if( row[idx+1] == 0 ) continue;
			size_t j=0;
			for(; j<nVar; ++j) {
				if(j == idx) continue;
				if( row[j+1] != 0 ) break;
			}
			if( j == nVar ) return true;
		}
		return false;
	}
};


/** Implements the %POLYInfer procedure. */
class POLYCheck : public IECheck
{
public:
	/** Constructor. */
	POLYCheck()
	: IECheck()
	{
	}

	/** Destructor. */
	virtual ~POLYCheck()
	{
	}

	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const
	{
		if( !feasible )
			return true;
		if( td1.typeSystem() > Type::UP2BT ) // parametric types
			return false;
		return td2.typeSystem() <= Type::UP2BT;
	}

	virtual TypeDecl *infer(const TypeDecl &typedecl, const string &varlist)
	{
		if( verbose::level > 0 ) {
			cout << "POLYCheck::infer called" << endl;
			cout << "Type declaration: " << typedecl << endl;
			cout << "Variables to infer:" << varlist << endl;
		}
		TypeDecl *v = LPCheck::infer(typedecl, varlist);
		// unsatisfiable constraint
		if( !feasible )
			return v;
		// check if candidate exists
		size_t sz = v->size();
		bool gonext = false;
		size_t n_cols = ls->getNcols();
		for(size_t i=0; i<sz; ++i)
			if( (v->getType(i) == BOX) && (ls->getColPos(v->getVarName(i)) < n_cols) )
				{
					gonext = true;
					break;
				}
		if( !gonext )
			return v;

		// move def-typed variables in higher dimensions, both in ls, polHomo and polNonHomo
		// this is equivalent to introduce explicit parameters
		// and equations x = a for x:! in typedecl.
		// return count of number of def-typed variables in typedecl
		size_t nparams = _arrange_def(typedecl);
		_infer(typedecl, nparams, v);

		if( verbose::level > 0 )
			cout << "Inferred Type declaration: " << v->to_string() << endl;
		return v;
	}


protected:

	/** Add a parameter for each variable. Needed to model Box_r. */
	virtual void _duplicatevars(LinearSystem *m)
	{
		size_t ncol = m->getNcols();
		for(size_t i=0; i<ncol; ++i)
			m->add_column(string("_par")+string_util::tostr(i));
	}

	/** Count no of defs in input type declaration. No actual rearrangement. */
	size_t _arrange_def(const TypeDecl &typedecl)
	{
		size_t n_cols = ls->getNcols();
		size_t sztd = typedecl.size();
		size_t def_typed = 0;
		for(size_t i=0; i<sztd; ++i) {
			const Type &type = typedecl.getType(i);
			if( (type.tp == DEF) && !type.is_p ) {
				const string &varname = typedecl.getVarName(i);
				size_t colpos = ls->getColPos(varname);
				if( colpos >= n_cols )
					continue;
				def_typed++;
			}
		}
		return def_typed;
	}

	void _infer(const TypeDecl &typedecl, size_t nparams, TypeDecl *v)
	{
		size_t n_cols = ls->getNcols();
		size_t sz = v->size();

		if( nparams == 0 ) {
			// No parameters, parpoly will not work on 0 param
			// reason directly on the non-parameterized polyhedron polNonHomo
			// NOTE: this case could be eliminated by setting one stub parameters 
			if( verbose::level > 0 ) {
				cout << "No variables typed as |r." << endl;
				cout << "Solving the non-parameterized polyhedron:" << endl;
				Polyhedron_Print(stdout,P_VALUE_FMT,polNonHomo);
			}
			for(size_t i=0; i<sz; ++i) {
				const Type &type = v->getType(i);
				if( !(type >= BOX) || (type == DEF) )
					continue;
				size_t idx = ls->getColPos(v->getVarName(i));
				if(idx >= n_cols) // x not in vars(c)
					continue;
				double range = poly_util::range_width(polNonHomo, idx); // compute range
				if( range >= 0 )
					v->setType(i, Type(range) );
			}
			return;
		}

		Param_Polyhedron *PA = _get_param_poly(typedecl, nparams, v, false);
		for(size_t i=0; i<sz; ++i) {
			const Type &type = v->getType(i);
			if( !(type >= BOX) || (type == DEF) )
				continue;
			size_t idx = ls->getColPos(v->getVarName(i));
			if( idx >= n_cols )
				continue; // variable not in c
			if(	check_generating(PA->Rays, idx) ) {
				double range = poly_util::range_width(PA->V, PA->D, idx);
				//cout << v->getVarName(i) << " range " << range << endl;
				if( range >= 0 )
					v->setType(i, Type(range) );
			}
		}
		Param_Polyhedron_Free( PA );
	}

	/** Build parameterized polyhedron for c and given typedeclaration. */
	Param_Polyhedron * _get_param_poly(const TypeDecl &typedecl, size_t nparams, 
								TypeDecl *v, bool parTypes)
	{
		// at least one parameter
		size_t n_cols = ls->getNcols();
		size_t tsz = typedecl.size();
		// counts number of rows in phi(typedecl) -- nparams is the number of columns!
		size_t defrows = 0;
		for(size_t i=0; i<tsz; ++i) {
			const Type &type = typedecl.getType(i);
			if( type.tp != DEF )
				continue;
			if( type.is_p && !parTypes)
				continue;
			const string &varname = typedecl.getVarName(i);
			size_t colpos = ls->getColPos(varname);
			if( colpos >= n_cols )
				continue;
			defrows += type.is_p ? 3 : (type.range == 0 ? 1 : 2); // ! needs 1 row, |r r > 0 needs 2 rows, |e needs 3 rows
		}
		// matrix for phi(typedecl)
		Matrix *MphiTD = Matrix_Alloc(defrows, n_cols+2);
		size_t j=0;
		size_t ppar = n_cols - nparams;
		size_t pos = 0;
		for(size_t i=0; i<tsz; ++i) {
			const Type &type = typedecl.getType(i);
			if( type.tp != DEF )
				continue;
			if( type.is_p && !parTypes)
				continue;
			const string &varname = typedecl.getVarName(i);
			size_t colpos = ls->getColPos(varname);
			if( colpos >= n_cols )
				continue;
			::Value *row = MphiTD->p[j++];
			size_t colposPar = ppar++; // use pool of parameters from below
			if( type.is_p ) {
				// set x <= a
				row[0] = 1;
				row[colpos+1] = 1;
				row[colposPar+1] = -1;
				row = MphiTD->p[j++];
				// set e >= 0
				::Value factor = set_par_table(pos++, row, n_cols);
				::Value *row2 = MphiTD->p[j++];
				// copy e >= 0
				memcpy(row2, row, sizeof(::Value)*(n_cols+2));
				// add a and x in e + a >= x
				row2[colpos+1] = -factor;
				row2[colposPar+1] = factor;
				continue;
			}
			//cout << varname << " " << colpos << " " << colposPar << endl;
			if( type.range == 0 ) {
				// set  x = a
				row[0] = 0;
				row[colpos+1] = 1;
				row[colposPar+1] = 1;
				continue;
			}
			// set  a <= x <= a + r
			row[0] = 1;
			row[colpos+1] = 1;
			row[colposPar+1] = -1;
			row = MphiTD->p[j++];
			double dec = string_util::decimals( type.range );
			row[0] = 1;
			row[colpos+1] = -(::Value)dec;
			row[colposPar+1] = (::Value)dec;
			row[n_cols+1] = (::Value)(type.range*dec);
		}

		if( verbose::level > 0 ) {
			cout << "  ";
			for(size_t i=0; i<n_cols; ++i) 
				cout << ls->getColName(i) << " ";
			cout << " | const" << endl;
			Matrix_Print(stdout,P_VALUE_FMT,MphiTD);
		}
		// polyhedron for phi(typedecl)
		Polyhedron *PphiTD = Constraints2Polyhedron(MphiTD, 0);
		Matrix_Free(MphiTD);
		
		if( PphiTD->NbRays == 0 ) { // not feasible parametric types
			Domain_Free(PphiTD);
			return NULL;
		}
		
		if( verbose::level > 0 )
			Polyhedron_Print(stdout,P_VALUE_FMT,PphiTD);
		// polyhedron for c, phi(typedecl)
		Polyhedron *pol = DomainIntersection(polNonHomo, PphiTD, 0);
		Domain_Free(PphiTD);
		// parameter empty matrix
		Matrix *Mpar = Matrix_Alloc(0, nparams+2);
		Polyhedron *Ppar = Constraints2Polyhedron(Mpar, 0);
		Matrix_Free(Mpar);
		Param_Polyhedron *PA = Polyhedron2Param_Vertices(pol, Ppar, 0);
		Domain_Free(Ppar);

		// Verbose output.
		if( verbose::level > 0 ) {
			  cout << "No of cols = " << n_cols << endl;
			  cout << "No of parameters = " << nparams << endl;
			  cout << "No of rows = " << defrows << endl;
			  char **param_name = new char*[nparams];
			  cout << "Parameters:" << endl;
			  ;
			  for(size_t i =0, basename = n_cols - nparams; i < nparams; ++i ) {
				  param_name[i] = (char *)ls->getColName(basename++).c_str();
				  cout << "  " << param_name[i] << endl;
			  }
			  cout << "Parameterized polyhedron:" << endl;
			  Polyhedron_Print(stdout,P_VALUE_FMT, (pol != NULL) ? pol : polNonHomo);
			  cout << "Parameterized vertices:" << endl;
			  Param_Vertices *PV = PA->V;
			  Param_Domain *PD = PA->D;
			  size_t vn = 0;
			  while(PV) {
				cout << "Vertex : " << ++vn << endl;
				Print_Vertex(stdout,PV->Vertex,(const char **)param_name);
				cout << "    If : " << endl;
				Print_Domain(stdout,PD->Domain,(const char **)param_name);
				Polyhedron_Print(stdout,P_VALUE_FMT,PD->Domain);
				PV = PV->next;
				PD = PD->next;
			 }
			 cout << "No vertices: " << vn << endl;
			 delete[] param_name;
		}

		Domain_Free(pol);
		return PA;
	}

	/** Stub method. */
	virtual ::Value set_par_table(size_t i, ::Value *row, size_t n_cols)
	{
		return 0;
	}

	/** Stub method. */
	virtual size_t get_alt_pos(const string &varname)
	{
		return 0;
	}

	/** Check if a variable is non-constant in a generating matrix. */
	static bool check_generating(Matrix *Rays, size_t idx)
	{
		size_t NbRays = Rays->NbRows;
		for (size_t i=0;i<NbRays;++i)
			if( Rays->p[i][idx+1] != 0 )
				return false;
		return true;
	}

};

/** Implements the %ParCheck procedure. */
class ParCheck : public POLYCheck
{
public:
	/** Constructor. */
	ParCheck()
	: POLYCheck(), parls(NULL), leftvars(NULL)
	{
	}

	/** Destructor. */
	virtual ~ParCheck()
	{
		if( parls != NULL )
			delete parls;
	}

	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const
	{
		return true;
	}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
	#define ERROR_PARNOLOCA "Syntax error: no local type variables admitted"
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

	virtual bool check(const TypeDecl &typedecl, const TypeDecl &td2, string &diff)
	{
		string varlist = td2.getVarList();
		TypeDecl *v = LPCheck::infer(typedecl, varlist);
		if( !feasible ) { 
			delete v;
			return true;
		}
		// check if inferred v is already implying td2
		size_t td2sz = td2.size();
		bool stop = true;
		for(size_t i=0; i<td2sz; ++i) {
			const Type &tp2 = td2.getType(i);
			size_t pos=0;
			v->find( td2.getVarName(i), pos );
			const Type &tp1 = v->getType(pos);
			// tp1 >= tp2   means v implies td2
			// tp1 >= BOX means LPCheck::infer does not proved invalidity
			if( tp1 >= tp2 )
				continue; // check ok
			stop = false;
			if( tp1 >= BOX )
				continue; // check ok
			diff = "expected " + td2.to_string() + " - actual: " + v->to_string();
			delete v;
			return false;
		}
		if( stop ) {
			delete v;
			return true;
		}
		size_t n_cols = ls->getNcols();
		parls = new LinearSystem(); // system of parametric constraints, only to compute set of parametric variables
		leftvars = new LinearSystem();
		size_t nparams = _arrange_def(typedecl);
		size_t newvars = leftvars->getNcols();
		parls->freeze();
		for(size_t i=0; i<newvars; ++i) 
			ls->renameColName(nvars_c+i, leftvars->getColName(i));
	
		if( nparams + newvars > n_cols - nvars_c )
			throw runtime_error("Out of parameter pool.");

		Param_Polyhedron *PA = _get_param_poly(typedecl, nparams, v, true);
		if( PA == NULL ) { // unfeasible typedecl
			for(size_t i=0; i<newvars; ++i) 
				ls->renameColName(nvars_c+i,string("_par")+string_util::tostr(i));
			delete v;
			delete parls; // delete linear system of parametric variables
			delete leftvars;
			parls = leftvars = NULL;
			return true;
		}
		bool result = true;
		for(size_t i=0; i<td2sz; ++i) {
			const Type &tp2 = td2.getType(i);
			if( tp2.tp != DEF )
				continue; // no check, LPCheck was complete
			size_t pos=0;
			v->find( td2.getVarName(i), pos );
			const Type &tp1 = v->getType(pos);
			if( tp1 >= tp2 )
				continue; // check ok
			//cout << td2.getVarName(i) << " checked." << endl;
			size_t idx = ls->getColPos(td2.getVarName(i));
			if( idx >= n_cols )
				throw runtime_error(ERROR_PARNOLOCA);
			if(	!check_generating(PA->Rays, idx) ) {
				//cout << td2.getVarName(i) << " failed generating." << endl;
				result = false; // inferred for tp1 is BOX, not more
				continue;
			}
			if( tp2.is_p ) {
				//cout << td2.getVarName(i) << " is parametric." << endl;
				::Value rowe[nparams+3];
				// r1[0], .., r1[d-1] variable coefficients
				// r1[d] alpha
				// r1[d+1] divisor
				try {
					parls->add(tp2.parametric+">=0");
				} catch( runtime_error e ) {
					throw runtime_error(ERROR_PARNOLOCA);
				}
				memset(rowe, 0, (nparams+3)*sizeof(::Value));
				rowe[nparams+2] = set_par_table(parls->getNrows()-1,rowe,nparams);
				//cout << *parls << endl;
				//for(size_t i=0; i<nparams+3; ++i)
				//	cout << "rowe[" << i << "]=" << rowe[i] << endl;
				// set rowe tp2.range
				if( !poly_util::range_width_check(PA->V, PA->D, idx, rowe+1) ) {
					v->setType(i, Type("|no_inference") );
					result = false;
				}
				continue;
			}
			// tp2 is DEF or BOX
			double range = poly_util::range_width(PA->V, PA->D, idx, tp2.range);
			if( (range < 0) || (range > tp2.range) ) {
				result = false;
				if( range > tp2.range )
					v->setType(i, Type(range) );
			}
		}
		Param_Polyhedron_Free( PA );
		if( !result )
			diff = "expected " + td2.to_string() + " - actual: " + v->to_string();
		for(size_t i=0; i<newvars; ++i) 
			ls->renameColName(nvars_c+i,string("_par")+string_util::tostr(i));
		delete v;
		delete parls; // delete linear system of parametric variables
		delete leftvars;
		parls = leftvars = NULL;
		return result;
	}

	/** Set  parameter pool size. */
	virtual void set_par_pool(size_t newmin, size_t newmax) {
		_parpool_min = newmin;
		_parpool_max = newmax;
	}
	
	/** Set  parameter pool size. */
	virtual void set_par_pool(const TypeDecl &typedecl, const string &constraint) {
		LinearSystem *lstemp = ls;
		ls = new LinearSystem(constraint);
		parls = new LinearSystem(); // system of parametric constraints, only to compute set of parametric variables
		leftvars = new LinearSystem();
		_parpool_min = _parpool_max = _arrange_def(typedecl) + leftvars->getNcols();
//		cout << "_parpool_min = _parpool_max= " << _parpool_min << endl;
		delete parls;
		delete leftvars;
		parls = leftvars = NULL;
		delete ls;
		ls = lstemp;
	}

protected:

	/** Number of default parameters for each variable */
	static size_t _dup_var_max;
	/** Number of total minimum parameters */
	static size_t _parpool_min;
	/** Number of total maximum parameters */
	static size_t _parpool_max;
	/** Linear system of parametric expressions. */
	LinearSystem *parls;
	/** Linear system of variables appearing in td1 but not in c. */
	LinearSystem *leftvars;
	
	/** Add a parameter for each variable. Needed to model Box_r. */
	virtual void _duplicatevars(LinearSystem *m)
	{
		size_t ncol = _dup_var_max*(1+nvars_c);
		if( ncol < _parpool_min )
			ncol = _parpool_min;
		else if( _parpool_max && (ncol > _parpool_max) )
			ncol = _parpool_max;

		for(size_t i=0; i<ncol; ++i)
			m->add_column(string("_par")+string_util::tostr(i));
	}

	/** Count no of defs in input type declaration. No actual rearrangement. */
	size_t _arrange_def(const TypeDecl &typedecl)
	{
		size_t n_cols = ls->getNcols();
		size_t sztd = typedecl.size();
		size_t def_typed = 0;
		for(size_t i=0; i<sztd; ++i) {
			const Type &type = typedecl.getType(i);
			if( type.tp == DEF ) {
				const string &varname = typedecl.getVarName(i);
				size_t colpos = ls->getColPos(varname);
				if( colpos >= n_cols )
					if( type.is_p )
						leftvars->add_column(varname);
					else
						continue;
				def_typed++;
				if( type.is_p )
					parls->add(type.parametric+">=0");
			}
		}
		def_typed += parls->getNcols(); // add parametric variables
		if( verbose::level > 0 )
			cout << endl << "Phi(td1) = " << endl << *parls;
		if( def_typed == 0 )
			def_typed = 1; //stub parameter, parameterized polyhedra need at least 1 parameter
		return def_typed;
	}

	virtual ::Value set_par_table(size_t i, ::Value *row, size_t n_cols)
	{
		size_t ncol = parls->getNcols();
		row[0] = (parls->getRowType(i)==LinearSystem::MAT_EQ) ? 0 : 1;
		double factor = 1;
		for(size_t j=0;j<=ncol;++j) {
			double dec = string_util::decimals( parls->get(i, j) );
			if( dec > factor )
				factor = dec;
		}
		size_t base = n_cols - ncol;
		for(size_t j=1;j<=ncol;++j) {
			row[base+j] = -(::Value)(parls->get(i,j-1)*factor);
		}
		row[base+ncol+1] = (::Value)(parls->get(i,ncol)*factor);
		return (::Value)factor;
	}

};

} // end of namespace clpt

#include <lp_lib.h>
namespace clpt
{
/** Implements the %LP2Check procedure. %LP2Check is a linear
	programming based version of LPCheck. Based on LPSolve:
	http://sourceforge.net/projects/lpsolve */
class LP2Check : public ISolve
{
public:
	/** Constructor. */
	LP2Check(const string &constraint = "")
	: ls(NULL), lp(NULL), row(NULL)
	{
		set(constraint);
	}

	/** Destructor. */
	virtual ~LP2Check()
	{
		clear();
	}

	virtual void set(const string &constraint)
	{
		clear();
		// convert constraint to a linear system
		ls = new LinearSystem(constraint);
		if( verbose::level > 0 ) {
			cout << "LP2Check called" << endl;
			cout << "Constraint: " << constraint << endl;
			cout << "Linear system:" << endl << *ls << endl;
		}
		// get variable list
		cvars = ls->getVarList();
		// convert linear system to LPsolve Matrix
		bool isStrict = _parse(ls);
		if( verbose::level > 0 ) {
			cout << "Linear programming problem:" << endl;
			print_lp(lp);
		}
		// check if feasible
		feasible = (ls->getNrows()==0) || (solve(lp) == OPTIMAL);
		if( feasible && isStrict ) {
			// check feasibility in the strict case
			REAL *ptr_pv;
			get_ptr_primal_solution(lp, &ptr_pv);
			if( *ptr_pv == 0 )
				feasible = false;
		}
		if( !feasible ) {
			// empty polyhedron
			if( verbose::level > 0 )
				cout << "Unfeasible linear programming problem." << endl;
			if( row != NULL ) {
				delete[] row;
				row = NULL;
			}
			delete_lp(lp);
			lp = NULL;
			delete ls;
			ls = NULL;
			feasible = false;
			return;
		}
		// make it nonstrict
		if( isStrict )
			del_column(lp, (int)ls->getNcols()+1);
		// make it homogeneous
		int nrows = (int)ls->getNrows();
		for(int i=1; i<=nrows; ++i)
			set_rh(lp, i, 0);
	}

	virtual bool isCFeasible() const
	{
		return feasible;
	}

	virtual string getCVarList() const
	{
		return cvars;
	}

	virtual bool is_check_complete(const TypeDecl &td1, const TypeDecl &td2) const
	{
		if( !feasible )
			return true;
		if( td1.typeSystem() > Type::UP2BT ) // parametric types
			return false;
		Type::TypeSystem ts2 = td2.typeSystem();
		if( ts2 <= Type::UP2BOX )
			return true;
		if( ts2 == Type::UP2BT )
			return ls->getNrows()==0;
		return false;
	}

	virtual TypeDecl *infer(const TypeDecl &typedecl, const string &varlist)
	{
		if( verbose::level > 0 ) {
			cout << "LP2Check::infer called" << endl;
			cout << "Type declaration: " << typedecl << endl;
			cout << "Variables to infer:" << varlist << endl;
		}

		// assign at least the types in typedecl
		TypeDecl *v = typedecl.restrict(varlist);
		// unsatisfiable constraint first
		if( !feasible ) {
			v->setAllType(DEF);
			if( verbose::level > 0 ) {
				cout << "Unsatisfiable constraint." << endl;
				cout << "Inferred type declaration: " << v->to_string() << endl;
			}
			return v;
		}
		size_t sz = v->size();
		size_t n_cols = ls->getNcols();
		// count variables that have to be typed
		size_t to_type = 0;
		for(size_t i=0; i<sz; ++i)
			if( !(v->getType(i) >= BOX) && // x : \tau in typedecl, with BOX >_t \tau
				(ls->getColPos(v->getVarName(i)) < n_cols) ) // x in vars(c)
				++to_type;
		if( to_type == 0 )
			return v;

		// set bounds for input type declaration
		int ncol = get_Ncolumns(lp);
		for(int i=1;i<=ncol;++i)
			set_unbounded(lp, i);
		size_t tdsz = typedecl.size();
		for(size_t i=0; i<tdsz; ++i) {
			int col = get_nameindex(lp, (char *)typedecl.getVarName(i).c_str(), 0);
//				cout << typedecl.getVarName(i) << " " << col << endl;
			if( col < 0 )
				continue;
			switch( typedecl.getType(i).tp ) {
				case DEF:
				case BOX:
					set_bounds(lp, col, 0, 0);
					break;
				case UP:
					set_upbo(lp, col, 0);
					break;
				case DOWN:
					set_lowbo(lp, col, 0);
					break;
				case STAR:
					break;
				default:
					break;
			}
		}

		if( verbose::level > 0 ) {
			cout << "Homogeneous linear programming problem:" << endl;
			print_lp(lp);
		}

		// for each variable to be typed
		for(size_t i=0; i<sz; ++i) {
			const Type &type = v->getType(i);
			if( !(type >= BOX) ) {
				size_t idx = ls->getColPos(v->getVarName(i));
				if(idx < n_cols)
					v->setType(i, _infer(type.tp, idx+1));
			}
		}

		if( verbose::level > 0 )
			cout << "Inferred Type declaration: " << v->to_string() << endl;
		return v;
	}

protected:
	/** Constraint in linear system form. */
	LinearSystem *ls;
	/** Linear programming problem (nonstrict). */
	lprec *lp;
	/** Is constraint satisfiable? */
	bool feasible;
	/** Constraint variables. */
	string cvars;
	/** Row in lpsolve format. */
	REAL *row;

	/** Clear data. */
	virtual void clear()
	{
		if( row != NULL ) {
			delete[] row;
			row = NULL;
		}
		if( lp != NULL ) {
			delete_lp(lp);
			lp = NULL;
		}
		if( ls != NULL ) {
			delete ls;
			ls = NULL;
		}
		feasible = true;
		cvars = "";
	}

	/** Parse a linear system into the lpsolve format. */
	bool _parse(LinearSystem *m)
	{
		bool strict = m->isStrict();
		int nrows = (int)m->getNrows();
		int ncol = (int)m->getNcols();
		lp=make_lp(0,ncol+ (strict ? 1 : 0)); // epsilon col
		if(lp == NULL)
			throw runtime_error("LP2Check::_parse error.");
		static char _dummy_name[] = "sample"; // needed to avoid warning on calling a set_lp_name( , char*)
		set_lp_name(lp, _dummy_name);
		set_verbose(lp, 0);
		row = new REAL[ncol+ (strict ? 2 : 1)];
		row[0] = 0;
		for(int i=0; i<nrows; ++i) {
			for(int j=0; j<ncol; ++j)
				row[j+1] = m->get(i, j);
			int rtype = (m->getRowType(i)==LinearSystem::MAT_EQ) ? ROWTYPE_EQ : ROWTYPE_LE;
			if( strict ) {
				// epsilon col
				if( m->getRowType(i)==LinearSystem::MAT_LT )
					row[ncol+1] = 1;
				else
					row[ncol+1] = 0;
			}
			add_constraint(lp, row, rtype, m->get(i, ncol));
		}
		for(int i=1; i<=ncol; ++i) {
			set_col_name(lp, i, (char *)m->getColName(i-1).c_str() );
			set_unbounded(lp, i);
		}
		for(int i=0; i<=ncol; ++i)
			row[i] = 0;
		if( strict ) {
			set_bounds(lp, ncol+1, 0, 1);
			row[ncol+1] = 1;
		}
		// obj
		set_obj_fn(lp, row);
		set_maxim(lp);
		return strict;
	}

	/** Infer the type by solving a linear programming problem. */
	Type _infer(TypeValues tp, size_t idx)
	{
		int status_up = OPTIMAL;
		if( tp != UP ) {
			row[idx] = 1;
			set_obj_fn(lp, row);
			status_up = solve(lp);
		}
		int status_down = OPTIMAL;
		if( tp != DOWN ) {
			row[idx] = -1;
			set_obj_fn(lp, row);
			status_down = solve(lp);
		}
		row[idx] = 0;
		if( status_up == OPTIMAL ) {
			if( status_down == OPTIMAL )
				return BOX;
			return UP;
		}
		if( status_down == OPTIMAL )
			return DOWN;
		return STAR;
	}
};

} // end of namespace clpt


using namespace clpt;

#endif // !defined(_POLY_H__INCLUDED_)
